
/*
 * Copyright 2007 Eric Sink
 * 
 * You may use this code under the terms of any of the
 * following licenses, your choice:
 * 
 * 1)  The GNU General Public License, Version 2
 *      http://www.opensource.org/licenses/gpl-license.php
 * 
 * 2)  The Apache License, Version 2.0
 *      http://www.opensource.org/licenses/apache2.0.php
 * 
 * 3)  The MIT License
 *      http://www.opensource.org/licenses/mit-license.php
 *
 * I am publishing this code on my blog as sample
 * code.  I am not intending to maintain this as a 
 * regular ongoing open source project or
 * anything like that.  I'm not looking for contributors
 * or hosting on sourceforge or anything like that.
 * 
 * Nonetheless, I hate it when I see an article with
 * sample code and it's not clear if I am allowed to
 * use the code or not.  The ambiguity is annoying.
 * So I am making this code available under your 
 * choice of open source licenses as described above.  
 * Informally and in a nutshell, you can use this code 
 * for any purpose as long as I am not liable for anything 
 * that goes wrong.  For a more tedious and formal 
 * explanation, pick one of the licenses above and use it.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;


namespace contagion
{
    /// <summary>
    /// Interaction logic for Window1.xaml
    /// </summary>

    public partial class Window1 : System.Windows.Window
    {
        private void fix_button(Button b)
        {
            string stag = (string)b.Tag;
            int cnum = int.Parse(stag);

            if (cnum == bs.CurrentColor)
            {
                b.IsEnabled = false;
                b.Visibility = Visibility.Hidden;
            }
            else
            {
                List<cell> infected = bs.PretendTurn(cnum);
                if (infected.Count == 0)
                {
                    b.IsEnabled = false;
                    b.Visibility = Visibility.Hidden;
                }
                else
                {
                    b.IsEnabled = true;
                    b.Visibility = Visibility.Visible;
                }
            }
        }

        private void fix_buttons()
        {
            fix_button(Button1);
            fix_button(Button2);
            fix_button(Button3);
            fix_button(Button4);
            fix_button(Button5);
            fix_button(Button6);
            fix_button(Button7);
        }

        private void update()
        {
            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    Polygon p = polys[brd.cells[c, r]];
                    p.Fill = GetBrush(bs.colors[c, r]);
                }
            }
            myStatus.Content = string.Format("{0} turns, {1} owned, {2} adjacent, {3} notadjacent", bs.Turns, bs.count_owned[0], bs.count_adjacent[0], bs.count_notadjacent[0]);
        }

        void do_reset()
        {
            bs = new boardstate(brd, seed);
        }

        void do_new()
        {
            seed = (int)(DateTime.Now.Ticks);
            do_reset();
        }

        void OnNew(object sender, RoutedEventArgs args)
        {
            do_new();
            update();
            fix_buttons();
        }

        void OnReset(object sender, RoutedEventArgs args)
        {
            do_reset();
            update();
            fix_buttons();
        }

        void PickColorGreedy(boardstate bs, out int cnum, out int max)
        {
            cnum = 0;
            max = 0;

            for (int i = 1; i <= 7; i++)
            {
                if (i != bs.CurrentColor)
                {
                    boardstate b2 = new boardstate(bs);
                    b2.DoTurn(i);
                    int count = b2.Score();
                    if (
                        (cnum == 0)
                        || (count > max)
                        )
                    {
                        cnum = i;
                        max = count;
                    }
                }
            }
        }

        private int FindMaxPlay(boardstate bs, int depth)
        {
            int cnum;
            int max;

            if (depth == 0)
            {
                PickColorGreedy(bs, out cnum, out max);
                return max;
            }

            max = 0;

            for (int i = 1; i <= 7; i++)
            {
                if (i != bs.CurrentColor)
                {
                    boardstate b2 = new boardstate(bs);
                    b2.DoTurn(i);
                    if (b2.IsDone())
                    {
                        return brd.ColCount*brd.RowCount + 100 * depth;
                    }

                    int result = FindMaxPlay(b2, depth - 1);
                    if (result == (brd.ColCount * brd.RowCount))
                    {
                        return result;
                    }

                    if (result > max)
                    {
                        max = result;
                    }
                }
            }
            return max;
        }

        private int PickColorWithLookahead(boardstate bs, int depth)
        {
            int cnum = 0;
            int max = 0;

            if (depth == 0)
            {
                PickColorGreedy(bs, out cnum, out max);
                return cnum;
            }

            for (int i = 1; i <= 7; i++)
            {
                if (i != bs.CurrentColor)
                {
                    boardstate b2 = new boardstate(bs);
                    b2.DoTurn(i);
                    if (b2.IsDone())
                    {
                        return i;
                    }

                    if (b2.count_owned[0] > bs.count_owned[0])
                    {
                        int result = FindMaxPlay(b2, depth - 1);

                        if (
                            (max == 0)
                            || (result > max)
                            )
                        {
                            cnum = i;
                            max = result;
                        }
                    }
                }
            }
            return cnum;
        }

        void OnBest(object sender, RoutedEventArgs args)
        {
            int cnum = PickColorWithLookahead(bs,3);
            bs.DoTurn(cnum);
            update();
            fix_buttons();
        }

        void OnLook(object sender, RoutedEventArgs args)
        {
            Button btn = sender as Button;
            string stag = (string)btn.Tag;
            int depth = int.Parse(stag);

            while (!bs.IsDone())
            {
                int cnum = PickColorWithLookahead(bs, depth);
                bs.DoTurn(cnum);
                update();
                Refresh(myCanvas);
            }
        }

        private int CountTurns(boardstate bs, int depth)
        {
            boardstate bmine = new boardstate(bs);
            while (!bmine.IsDone())
            {
                int cnum = PickColorWithLookahead(bmine, depth);
                bmine.DoTurn(cnum);
            }
            return bmine.Turns;
        }

        private int CountTurnsGreedy(boardstate bs)
        {
            boardstate bmine = new boardstate(bs);
            while (!bmine.IsDone())
            {
                int cnum;
                int max;
                PickColorGreedy(bmine, out cnum, out max);
                bmine.DoTurn(cnum);
            }
            return bmine.Turns;
        }

        public class stats
        {
            public int min = int.MaxValue;
            public int max = int.MinValue;
            public int total = 0;
            public int count = 0;
        }

        private Dictionary<int, stats> results = new Dictionary<int, stats>();

        private void oneboard_onedepth(boardstate bs, int depth)
        {
            int turns;

            if (depth == 0)
            {
                turns = CountTurnsGreedy(bs);
            }
            else
            {
                turns = CountTurns(bs, depth);
            }

            if (!results.ContainsKey(depth))
            {
                results[depth] = new stats();
            }

            stats st = results[depth];
            st.count++;
            st.total += turns;
            st.min = Math.Min(st.min, turns);
            st.max = Math.Max(st.max, turns);
        }

        private void oneboard(boardstate bs)
        {
            for (int i = 0; i < 6; i++)
            {
                oneboard_onedepth(bs, i);
            }
        }

        private void dump()
        {
            foreach (int depth in results.Keys)
            {
                stats st = results[depth];
                Console.WriteLine("Depth {0}:  Min={1}  Max={2}  Avg={3}", depth, st.min, st.max, st.total / st.count);
            }
        }

        private void do_stats()
        {
            for (int i = 0; i < 100; i++)
            {
                boardstate bs = new boardstate(brd);
                oneboard(bs);
            }
            dump();
        }

        private delegate void NoArgDelegate();
        public static void Refresh(DependencyObject obj)
        {
            obj.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.ApplicationIdle,
                (NoArgDelegate)delegate
            {
            });
        }

        void OnGreedy(object sender, RoutedEventArgs args)
        {
            while (!bs.IsDone())
            {
                int cnum;
                int max;
                PickColorGreedy(bs, out cnum, out max);
                bs.DoTurn(cnum);
                update();
                Refresh(myCanvas);
            }
        }

        void OnClick(object sender, RoutedEventArgs args)
        {
            Button btn = sender as Button;
            string stag = (string)btn.Tag;
            int cnum = int.Parse(stag);
            bs.DoTurn(cnum);

            update();
            fix_buttons();
        }

        private Brush GetBrush(int cnum)
        {
            switch (cnum)
            {
                case 1: return Brushes.Blue;
                case 2: return Brushes.Red;
                case 3: return Brushes.Purple;
                case 4: return Brushes.Yellow;
                case 5: return Brushes.Pink;
                case 6: return Brushes.Orange;
                case 7: return Brushes.Green;
            }
            throw new Exception();
        }

        board brd = new board(20, 16);
        boardstate bs;
        Dictionary<cell, Polygon> polys = new Dictionary<cell, Polygon>();
        int seed;

        public Window1()
        {
            InitializeComponent();

            // do_stats();

            do_new();

            const int size = 16;

            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    cell ce = brd.cells[c, r];
                    Polygon p = new Polygon();

                    int off = 1 - (c % 2);
                    PointCollection pc = new PointCollection();
                    pc.Add(new Point(size + (c * 2 + 0) * size, size + (off + r * 2 + 1) * size));
                    pc.Add(new Point(size + (c * 2 + 1) * size, size + (off + r * 2 + 0) * size));
                    pc.Add(new Point(size + (c * 2 + 2) * size, size + (off + r * 2 + 0) * size));
                    pc.Add(new Point(size + (c * 2 + 3) * size, size + (off + r * 2 + 1) * size));
                    pc.Add(new Point(size + (c * 2 + 2) * size, size + (off + r * 2 + 2) * size));
                    pc.Add(new Point(size + (c * 2 + 1) * size, size + (off + r * 2 + 2) * size));
                    p.Points = pc;

                    p.Fill = GetBrush(bs.colors[c, r]);

                    p.Tag = ce;

                    polys[ce] = p;

                    myCanvas.Children.Add(p);
                }
            }

            update();
            fix_buttons();
        }

    }
}