
/*
 * Copyright 2007 Eric Sink
 * 
 * You may use this code under the terms of any of the
 * following licenses, your choice:
 * 
 * 1)  The GNU General Public License, Version 2
 *      http://www.opensource.org/licenses/gpl-license.php
 * 
 * 2)  The Apache License, Version 2.0
 *      http://www.opensource.org/licenses/apache2.0.php
 * 
 * 3)  The MIT License
 *      http://www.opensource.org/licenses/mit-license.php
 *
 * I am publishing this code on my blog as sample
 * code.  I am not intending to maintain this as a 
 * regular ongoing open source project or
 * anything like that.  I'm not looking for contributors
 * or hosting on sourceforge or anything like that.
 * 
 * Nonetheless, I hate it when I see an article with
 * sample code and it's not clear if I am allowed to
 * use the code or not.  The ambiguity is annoying.
 * So I am making this code available under your 
 * choice of open source licenses as described above.  
 * Informally and in a nutshell, you can use this code 
 * for any purpose as long as I am not liable for anything 
 * that goes wrong.  For a more tedious and formal 
 * explanation, pick one of the licenses above and use it.
 * 
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;

namespace contagion
{
    public class boardstate
    {
        public int Turns = 0;
        public int CurrentColor;
        public board brd;
        public int[,] colors;
        public bool[,] owned;
        public int[] count_owned = new int[8];
        public int[] count_adjacent = new int[8];
        public int[] count_notadjacent = new int[8];

        public int Score()
        {
            return count_owned[0] + brd.RowCount * brd.ColCount - count_notadjacent[0];
            //return count_owned[0];
        }

        private void getNeighbors(cell ce, int clr, List<cell> adj)
        {
            ce.ApplyToNeighbors(
                delegate(cell cn)
                {
                    if (
                        (!adj.Contains(cn))
                        && (!owned[cn.col, cn.row])
                        )
                    {
                        if (
                            (clr == 0)
                            || (clr == colors[cn.col, cn.row])
                            )
                        {
                            adj.Add(cn);
                            getNeighbors(cn, colors[cn.col, cn.row], adj);
                        }
                    }
                }
                );
        }

        private void update_count_adjacent()
        {
            List<cell> adj = new List<cell>();

            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    if (owned[c, r])
                    {
                        getNeighbors(brd.cells[c, r], 0, adj);
                    }
                }
            }

            for (int i = 0; i <= 7; i++)
            {
                count_adjacent[i] = 0;
            }

            foreach (cell ce in adj)
            {
                count_adjacent[0]++;
                count_adjacent[colors[ce.col, ce.row]]++;
            }

            for (int i = 0; i <= 7; i++)
            {
                count_notadjacent[i] = 0;
            }

            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    if (!owned[c, r])
                    {
                        if (!adj.Contains(brd.cells[c, r]))
                        {
                            count_notadjacent[0]++;
                            count_notadjacent[colors[c, r]]++;
                        }
                    }
                }
            }
        }

        private void update_counts()
        {
            update_count_owned();
            update_count_adjacent();
        }

        private void update_count_owned()
        {
            for (int i = 0; i <= 7; i++)
            {
                count_owned[i] = 0;
            }

            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    if (owned[c, r])
                    {
                        count_owned[0]++;
                        count_owned[colors[c, r]]++;
                    }
                }
            }
        }

        public bool IsDone()
        {
            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    if (!owned[c, r])
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        private int GetColor(cell ce)
        {
            return colors[ce.col, ce.row];
        }

        private void spread(cell ce, int cnum, List<cell> infected)
        {
            ce.ApplyToNeighbors(delegate(cell cn)
            {
                if (
                    (GetColor(cn) == cnum)
                    && (!infected.Contains(cn))
                    && (!owned[cn.col, cn.row])
                    )
                {
                    infected.Add(cn);
                    spread(cn, cnum, infected);
                }
            }
            );
        }

        private List<cell> spread(int cnum)
        {
            List<cell> infected = new List<cell>();
            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    if (owned[c, r])
                    {
                        spread(brd.cells[c, r], cnum, infected);
                    }
                }
            }
            return infected;
        }

        private void change(int cnum)
        {
            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    if (owned[c, r])
                    {
                        colors[c, r] = cnum;
                    }
                }
            }
        }

        public void DoTurn(int cnum)
        {
            change(cnum);
            List<cell> infected = spread(cnum);
            foreach (cell ce in infected)
            {
                colors[ce.col, ce.row] = cnum;
                owned[ce.col, ce.row] = true;
            }
            CurrentColor = cnum;
            Turns++;
            update_counts();
        }

        public List<cell> PretendTurn(int cnum)
        {
            List<cell> infected = spread(cnum);
            return infected;
        }

        private void createArrays()
        {
            colors = new int[brd.ColCount, brd.RowCount];
            owned = new bool[brd.ColCount, brd.RowCount];
        }

        private void init(board b, int seed)
        {
            brd = b;
            createArrays();

            Random rand = new Random(seed);
            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    colors[c, r] = rand.Next(1, 7);
                    owned[c, r] = ((c == 0) && (r == 0));
                }
            }
            CurrentColor = colors[0, 0];
            List<cell> infected = spread(CurrentColor);
            foreach (cell ce in infected)
            {
                colors[ce.col, ce.row] = CurrentColor;
                owned[ce.col, ce.row] = true;
            }
            update_counts();
        }

        public boardstate(boardstate copy)
        {
            brd = copy.brd;
            createArrays();
            Turns = copy.Turns;

            for (int c = 0; c < brd.ColCount; c++)
            {
                for (int r = 0; r < brd.RowCount; r++)
                {
                    colors[c, r] = copy.colors[c, r];
                    owned[c, r] = copy.owned[c, r];
                }
            }
        }

        public boardstate(board b, int seed)
        {
            init(b, seed);
        }

        public boardstate(board b)
        {
            init(b, (int) (DateTime.Now.Ticks));
        }
    }

    public class board
    {
        public cell[,] cells;
        public int RowCount;
        public int ColCount;

        public board(int cols, int rows)
        {
            RowCount = rows;
            ColCount = cols;
            cells = new cell[ColCount, RowCount];
            for (int c = 0; c < ColCount; c++)
            {
                for (int r = 0; r < RowCount; r++)
                {
                    cell ce = new cell(this, c, r);
                    cells[c, r] = ce;
                }
            }
        }
    }

    public class cell
    {
        public board brd;
        public int col;
        public int row;

        public cell(board b, int c, int r)
        {
            brd = b;
            col = c;
            row = r;
        }

        public delegate void CellFunc(cell ce);

        public void ApplyToNeighbors(CellFunc cf)
        {
            // the cell just below me
            if (row > 0)
            {
                cf(brd.cells[col, row - 1]);
            }

            // the cell just above me
            if (row < (brd.RowCount-1))
            {
                cf(brd.cells[col, row + 1]);
            }

            // calculate the row of the hi and lo neighbors left and right
            int hi = row + 1 - (col % 2);
            int lo = hi - 1;

            // the cells to the left of me
            if (col > 0)
            {
                if (hi <= (brd.RowCount - 1))
                {
                    cf(brd.cells[col - 1, hi]);
                }
                if (lo >= 0)
                {
                    cf(brd.cells[col - 1, lo]);
                }
            }

            // the cells to the right of me
            if (col < (brd.ColCount-1))
            {
                if (hi <= (brd.RowCount - 1))
                {
                    cf(brd.cells[col + 1, hi]);
                }
                if (lo >= 0)
                {
                    cf(brd.cells[col + 1, lo]);
                }
            }
        }
    }
}
